#include <sys/kos.h>
#include <stdio.h>
#include "libkoremotor.h"

/* Knet address for the KoreMotor */
#define KMOT_ADDR 6



/******************************************************************
 * Write a 8bit value to the given i2c register for the given motor
 * mot  : motor
 * addr : register address
 * data : 8bit data 
 */
int kmot_I2cWrite(int8 mot, int8 addr, int8 data) 
{
    char send[6];
    char recv[1];
    int  status;

    send[0] = KMOT_ADDR;
    send[1] = 4;
    send[2] = 'W';
    send[3] = mot;
    send[4] = addr;
    send[5] = data;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,6,recv,1,2);
    msg_release_channel(0);
	
    return status;
}

/******************************************************************
 * Read a 8bit value from the given register for the given motor
 * Return the register value 
 */
int kmot_I2cRead(int8 mot, int8 addr)
{
    char send[5];
    char recv[2];
    int  status;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    if(status < 0)
	return status;
    else
	return recv[1];
}

/******************************************************************
 * Test the KNet link to the KoreMotor
 * return 1 if the link is active
 * return 0 if the link is down
 */
int kmot_TestLink()
{
    char send[3];
    char recv[2];
    int  status;

    send[0] = KMOT_ADDR;
    send[1] = 1;
    send[2] = 'Z';

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,3,recv,2,2);
    msg_release_channel(0);
	
    if(status<0 || recv[0]!='z' || recv[1]!=0)
	return 0;
    else	
	return 1;

}

/******************************************************************
 * Reset the error register for the given motor, all controls are
 * disabled while any error is active.
 */
int kmot_ResetError(int8 mot)
{
    return kmot_I2cWrite(mot,0x2D,0);
}

/******************************************************************
 * Set a target point for the given controler and the given motor.
 * mot    : motor
 * regtype : regulation
 *           reg_open = open loop
 *           reg_pos  = position control
 *           reg_posprofile = position control with a trapezoidal speed profile
 *           reg_speed   = speed regulation
 *           reg_current = current (torque) regulation 
 * setpoint : target point          
 */
int kmot_SetPoint(int8 mot, RegType reg, int32 setpoint)
{
    char send[9];
    char recv[1];
    int  status;

    send[0] = KMOT_ADDR;
    send[1] = 7;
    send[2] = 'P';
    send[3] = mot;
    send[4] = reg;
    send[5] = setpoint >> 24;
    send[6] = setpoint >> 16;
    send[7] = setpoint >> 8;
    send[8] = setpoint;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,9,recv,1,2);
    msg_release_channel(0);

    return status;
}

/******************************************************************
 * Read a given measure from the given motor. The value is returned 
 * in the value parameter.
 * mot : motor
 * mes : measurement type
 *       mes_pos     = position
 *       mes_speed   = speed
 *       mes_current = current
 * value : pointer to the returned value storage      
 */
int kmot_GetMeasure(int8 mot, Measure mes, int32* value)
{
    char send[5];
    char recv[5];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'M';
    send[3] = mot;
    send[4] = mes;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,5,2);
    msg_release_channel(0);

    *value  = recv[4] & 0x000000FF;
    *value += (recv[3] << 8)  & 0x0000FF00;
    *value += (recv[2] << 16) & 0x00FF0000;
    *value += (recv[1] << 24) & 0xFF000000;

    return status;
}

/******************************************************************
 * Set the limits for the given regulation. The exact behavior when 
 * a measure reaches a limit depends on the controler options. 
 */ 
int kmot_SetPositionLimit(int8 mot, int32 min, int32 max)
{
    char send[13];
    char recv[1];
    int  status;

    send[0]  = KMOT_ADDR;
    send[1]  = 11;
    send[2]  = 'L';
    send[3]  = mot;
    send[4]  = reg_pos;
    send[5]  = min >> 24;
    send[6]  = min >> 16;
    send[7]  = min >> 8;
    send[8]  = min;
    send[9]  = max >> 24;
    send[10] = max >> 16;
    send[11] = max >> 8;
    send[12] = max ;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,13,recv,1,2);
    msg_release_channel(0);
    

    return status;
}

int kmot_SetCurrentLimit(int8 mot, int32 max)
{
    char send[13];
    char recv[1];
    int  status;

    send[0]  = KMOT_ADDR;
    send[1]  = 11;
    send[2]  = 'L';
    send[3]  = mot;
    send[4]  = reg_current;
    send[5]  = max >> 24;
    send[6]  = max >> 16;
    send[7]  = max >> 8;
    send[8]  = max;
    send[9]  = 0;
    send[10] = 0;
    send[11] = 0;
    send[12] = 0;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,13,recv,1,2);
    msg_release_channel(0);
    

    return status;
}

/******************************************************************
 * Configure the PID controler for the given regulation
 * Please refer to kmot_SetPoint for Regtype definition.
 */
int kmot_ConfigurePID(int8 mot, RegType reg, int16 Kp, int16 Kd, int16 Ki)
{
    char send[11];
    char recv[1];
    int  status;

    send[0]  = KMOT_ADDR;
    send[1]  = 9;
    send[2]  = 'C';
    send[3]  = mot;
    send[4]  = reg;
    send[5]  = Kp >> 8;
    send[6]  = Kp;
    send[7]  = Kd >> 8;
    send[8]  = Kd;
    send[9]  = Ki >> 8;
    send[10] = Ki;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,11,recv,1,2);
    msg_release_channel(0);
    
    return status;
}

/******************************************************************
 * Read 32bit from the given motor at the given address. The read
 * bytes are:
 * LL byte = addr
 * LH byte = addr+1
 * HL byte = addr+2
 * HH byte = addr+3 
 * The value is returned in the value parameter.
 */
int kmot_I2cRead32(int8 mot, int8 addr, int32* value)
{
    char send[5];
    char recv[2];
    int  status;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    *value = recv[1] & 0x000000FF;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr+1;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    *value += (recv[1] << 8) & 0x0000FF00;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr+2;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    *value += (recv[1] << 16) & 0x00FF0000;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr+3;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    *value += (recv[1] << 24) & 0xFF000000;

    return status;
}

/******************************************************************
 * Read 16bit from the given motor at the given address. The Low byte
 * is the addr, the high byte is the addr+1. The value is returned
 * in the value parameter.
 */
int kmot_I2cRead16(int8 mot, int8 addr, int16* value)
{
    char send[5];
    char recv[2];
    int  status;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    *value = recv[1] & 0x000000FF;

    send[0] = KMOT_ADDR;
    send[1] = 3;
    send[2] = 'R';
    send[3] = mot;
    send[4] = addr+1;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,5,recv,2,2);
    msg_release_channel(0);

    *value += (recv[1] << 8) & 0x0000FF00;

    return status;
}

/******************************************************************
 * Save current configuration for the given controler to the controler
 * EEPROM. This EEPROM is not erased when the system is shut down.
 */
int kmot_SaveConfig(int8 mot)
{
    char send[4];
    char recv[1];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 2;
    send[2] = 'S';
    send[3] = mot;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,4,recv,1,2);
    msg_release_channel(0);

    /* wait for the EEPROM write completion */
    while(kmot_I2cRead(mot,0x28))
	tim_switch_fast();

    return status;
}

/******************************************************************
 * Set the position counter to the given value for the given motor.
 */
int kmot_SetPosition(int8 mot, int32 position)
{
    char send[8];
    char recv[1];
    int  status;

    send[0]  = KMOT_ADDR;
    send[1]  = 6;
    send[2]  = 'A';
    send[3]  = mot;
    send[4]  = position >> 24;
    send[5]  = position >> 16;
    send[6]  = position >> 8;
    send[7]  = position;

    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,8,recv,1,2);
    msg_release_channel(0);

    return status;
}
/******************************************************************
 * Set the profile for position regulation (reg_posprofile). 
 * The trapezoidal speed profile is calculated using the acceleration
 * until the maxspeed is reached. -acceleration is used to decelerate 
 * until target position is reached.
 */
int kmot_SetSpeedProfile(int8 mot, int16 maxspeed, int8 acceleration)
{
    char send[7];
    char recv[1];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 5;
    send[2] = 'J';
    send[3] = mot;
    send[4] = maxspeed >> 8;
    send[5] = maxspeed;
    send[6] = acceleration;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,7,recv,1,2);
    msg_release_channel(0);

    return status;
}

/******************************************************************
 * Search the system mechanical limits. The blocked time can adjust 
 * the test sensitivity and the current limit used for blocked 
 * detection is the one set with kmot_SetLimits. The given speed is 
 * used to perform the test. The min and max positions are retruned
 * with the minpos and maxpos parameters.
 * The speed regulator must be properly configured before calling the
 * search limit routine.
 */
int kmot_SearchLimits(int8 mot,int8 blocked_time,int32 speed,int32* minpos,int32* maxpos)
{
    char send[9];
    char recv[9];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 7;
    send[2] = 'N';
    send[3] = mot;
    send[4] = blocked_time;
    send[5]  = speed >> 24;
    send[6]  = speed >> 16;
    send[7]  = speed >> 8;
    send[8]  = speed;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,9,recv,1,2);
    msg_release_channel(0);

    while(kmot_I2cRead(mot,0x28))
	tim_switch_fast(); 

    kmot_I2cRead32(mot,0x4A,minpos);
    kmot_I2cRead32(mot,0x4E,maxpos);

    return status;
}

/******************************************************************
 * Select a source for the target point. Usually the setpoint is 
 * given by the application with knet commnands, on some cases, an
 * internal signal generator can provide a variable setpoint.
 * Source:
 * src_ext = external source
 * src_sqr = internal square signal generator
 * src_tri = internal triangle signal
 * src_sin = internal sinus signal (Not Implemented)
 */
int kmot_SetPointSource(int8 mot, RegType reg, SetPointSource src,
	int8 period, int8 amplitude, int8 offset)
{
    char send[9];
    char recv[1];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 7;
    send[2] = 'H';
    send[3] = mot;
    send[4] = reg;
    send[5] = src;
    send[6] = period;
    send[7] = amplitude;
    send[8] = offset;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,4,recv,1,2);
    msg_release_channel(0);

    return status;

}

/******************************************************************
 * Get the status flag and error flag registers. The restults are
 * returned within the two paramters.
 *  Error Flags:
 *  bit0 : Sample time too small
 *  bit1 : Watchdog timer overflow
 *  bit2 : Brown-out
 *  bit3 : Software stoped motor (if softstop enabled)
 *  bit4 : Motor blocked (if motorblock enabled)
 *  bit5 : Position out of range
 *  bit6 : Speed out of range
 *  bit7 : Torque out of range
 *
 *  Status flags:
 *  bit0 : Movement detected
 *  bit1 : Direction 0=negative 1=positive
 *  bit2 : On setpoint
 *  bit3 : Near setpoint
 *  bit4 : Command saturated
 *  bit5 : Antireset windup active
 *  bit6 : Software current control active
 *  bit7 : Softstop active
 */ 
#define STA_ON_SETPOINT   4
#define STA_NEAR_SETPOINT 8

int kmot_GetStatus(int8 mot, int8* statusflag, int8* errorflag)
{
    char send[4];
    char recv[3];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 2;
    send[2] = 'E';
    send[3] = mot;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,4,recv,3,2);
    msg_release_channel(0);

    *statusflag = recv[1];
    *errorflag  = recv[2]; 

    return status;
}

/******************************************************************
 * Set the options for the given controler. Each SW and HW option
 * parameter is a OR field of option flags.
 * 
 * sw_separated  : use alternate algorithm PID derivation
 * sw_windup     : activate the anti reset windup routine
 * sw_stopmin    : stop the motor if the min position is reached
 * sw_stopmax    : stop the motor if the max position is reached
 * sw_stop_error : generate an error when position is out of limits
 *                 in this case the error must be rested before any 
 *                 further commands can be executed
 * sw_blocked_stop : stop the motor if the blocked condition is met
 *                   (No movement, and current over the limit for
 *                   the blocked time period).
 * sw_current_ctrl : activate software current limitation (Not implemented)                       
 * sw_dir_inv      : invert the motor direction
 *
 * hw_startup    : startup mode (0 = idle mode, 1 = control mode)
 * hw_analog_set : use analog input for setpoint (Not Implemented)
 * hw_led        : Not Implemented
 * hw_resolution : resolution for the encoder (0 = 100%, 1 = 25%)
 * hw_torque_inv : invert the internal current measurement    
 * hw_opt1 : Not Implemented
 * hw_opt2 : Not Implemented
 * hw_opt3 : Not Implemented
 */

int kmot_SetOptions(int8 mot, int8 hwoption, int8 swoption)
{
    char send[6];
    char recv[1];
    int status;

    send[0] = KMOT_ADDR;
    send[1] = 4;
    send[2] = 'O';
    send[3] = mot;
    send[4] = hwoption;
    send[5] = swoption;
    
    while(msg_reserve_channel(0))
	tim_switch_fast();
    status = msg_snd_rec_message(send,6,recv,1,2);
    msg_release_channel(0);

    return status;
}

/******************************************************************/
int kmot_SetBlockedTime(int8 mot,int8 time)
{
    return kmot_I2cWrite(mot,0x46,time);
}

int kmot_SetSampleTime(int8 mot,int8 sample) 
{
    return kmot_I2cWrite(mot,0x45,sample);
}

int kmot_SetFilterOrder(int8 mot,int8 order) 
{
    return kmot_I2cWrite(mot,0x33,order);
}

int kmot_ResetController(int8 mot)
{
    return kmot_I2cWrite(mot,0x28,4);
}
